package reviewdog

import (
	"errors"
	"fmt"
	"sync"
)

// ResultMap represents a concurrent-safe map to store CheckResults generated by concurrent jobs.
type ResultMap struct {
	sm sync.Map
}

type Result struct {
	Level        string
	CheckResults []*CheckResult
}

// Store saves a new *Result into ResultMap.
func (rm *ResultMap) Store(key string, r *Result) {
	rm.sm.Store(key, r)
}

// Load fetches *Result from ResultMap
func (rm *ResultMap) Load(key string) (*Result, error) {
	v, ok := rm.sm.Load(key)
	if !ok {
		return nil, fmt.Errorf("fail to get the value of key %q from results", key)
	}

	t, ok := v.(*Result)
	if !ok {
		return nil, errors.New("stored type in ResultMap is invalid")
	}

	return t, nil
}

// Range retrieves `key` and `values` from ResultMap iteratively.
func (rm *ResultMap) Range(f func(key string, val *Result)) {
	rm.sm.Range(func(k, v interface{}) bool {
		f(k.(string), v.(*Result))
		return true
	})
}

// Len returns the length of ResultMap count. Len() is not yet officially not supported by Go. (ref: https://github.com/golang/go/issues/20680)
func (rm *ResultMap) Len() int {
	l := 0
	rm.sm.Range(func(_, _ interface{}) bool {
		l++
		return true
	})
	return l
}

// FilteredCheckMap represents a concurrent-safe map to store CheckResults generated by concurrent jobs.
type FilteredCheckMap struct {
	sm sync.Map
}

// Store saves a new []*FilteredCheck into FilteredCheckMap.
func (rm *FilteredCheckMap) Store(key string, r []*FilteredCheck) {
	rm.sm.Store(key, r)
}

// Load fetches []*FilteredCheck from FilteredCheckMap
func (rm *FilteredCheckMap) Load(key string) ([]*FilteredCheck, error) {
	v, ok := rm.sm.Load(key)
	if !ok {
		return nil, fmt.Errorf("fail to get the value of key %q from results", key)
	}

	t, ok := v.([]*FilteredCheck)
	if !ok {
		return nil, errors.New("stored type in FilteredCheckMap is invalid")
	}

	return t, nil
}

// Range retrieves `key` and `values` from FilteredCheckMap iteratively.
func (rm *FilteredCheckMap) Range(f func(key string, val []*FilteredCheck)) {
	rm.sm.Range(func(k, v interface{}) bool {
		f(k.(string), v.([]*FilteredCheck))
		return true
	})
}

// Len returns the length of FilteredCheckMap count. Len() is not yet officially not supported by Go. (ref: https://github.com/golang/go/issues/20680)
func (rm *FilteredCheckMap) Len() int {
	l := 0
	rm.sm.Range(func(_, _ interface{}) bool {
		l++
		return true
	})
	return l
}
